from django.apps import apps
from django.core.management import CommandError
from django.core.management.commands.squashmigrations import Command as Migrate

from data_migration.services.squasher import MigrationSquash


class Command(Migrate):
    """Extended `squashmigrations` management command."""

    def add_arguments(self, parser):  # noqa D102
        parser.add_argument(
            '--extract-data-migrations',
            dest='extract_data_migrations',
            action='store_true',
            help='Minimize current migration tree and emplace data migrations',
        )
        parser.add_argument(
            '--dry-run',
            dest='dry_run',
            action='store_true',
            help='Run squashing dry, does not create data_migration files.',
        )
        super().add_arguments(parser)

    def handle(self, **options):  # noqa D102
        app_label = options['app_label']
        extract_data_migrations = options['extract_data_migrations']
        # Validate app_label.
        try:
            apps.get_app_config(app_label)
        except LookupError as err:
            raise CommandError(str(err))

        if extract_data_migrations:
            squasher = MigrationSquash(app_label, options.get('dry_run'))
            squasher.squash()
            return squasher.log.log
        else:
            return super().handle(**options)



"""public helpers of package."""
import os


def get_package_version_string():
    """:return: version string of package."""
    name = 'data_migration'
    version = os.getenv('VERSION', '0.0.1a')
    return f'{name} {version}'



from django.core.management.commands.makemigrations import Command as Migrate


class Command(Migrate):
    """Alias for default `makemigrations` command provided by Django."""

    def handle(self, *args, **options):  # noqa D102
        return super().handle(*args, **options)



from django.core.management import BaseCommand

from data_migration.services.squasher import MigrationSquash
from data_migration.settings import internal_settings as data_migration_settings


class Command(BaseCommand):
    """
    Extended migrate command.

    Allows forward and backward migration of data/regular migrations
    """

    def add_arguments(self, parser):  # noqa D102
        parser.add_argument(
            '--app_labels', nargs='?', dest='app_labels',
            help='App label of an application to synchronize the state.',
        )
        parser.add_argument(
            '--all', '-a', action='store_true', dest='squash_all',
            help='Squash all apps.',
        )
        super().add_arguments(parser)

    def handle(self, *args, **options):  # noqa D102
        # extract parameters
        apps_to_squash = options.get('app_labels')
        if apps_to_squash is None and options['squash_all']:
            apps_to_squash = data_migration_settings.SQUASHABLE_APPS

        MigrationSquash(apps_to_squash).squash()



from typing import Dict, Optional

from django.core.signals import setting_changed

DATA_MIGRATION_DEFAULTS = {
    "SQUASHABLE_APPS": []
}


class DataMigrationSettings(object):
    def __init__(self, user_settings=None, defaults: Optional[Dict] = None):
        if defaults is None:
            defaults = DATA_MIGRATION_DEFAULTS

        self.settings = defaults.copy()
        if user_settings:
            self.update(user_settings)

    def update(self, settings):
        try:
            self.settings.update(getattr(settings, 'DATA_MIGRATION'))
        except AttributeError:
            self.settings.update(settings.get('DATA_MIGRATION', {}))

    def reload(self, settings):
        try:
            _user_settings = getattr(settings, 'DATA_MIGRATION')
            self.settings = _user_settings
        except AttributeError:
            pass

    def __getattr__(self, item):
        return self.settings[item]


internal_settings = DataMigrationSettings(None)


def reload(sender, setting, value, *args, **kwargs):
    if setting == 'DATA_MIGRATION':
        internal_settings.update(value)


setting_changed.connect(reload)



import os
from importlib import import_module
from types import FunctionType
from typing import Optional, List

from django.apps import apps
from django.db import connections
from django.db.migrations.exceptions import NodeNotFoundError
from django.db.migrations.executor import MigrationExecutor
from django.db.migrations.recorder import MigrationRecorder

from data_migration.services.node import Node

FunList = List[FunctionType]


class GraphNode:
    """Migration node representation within a graph."""

    def __init__(self, app_name: str, name: str,
                 dependencies: List[str],
                 migration_dependencies: List[str],
                 routines: FunList) -> None:
        self.routines = routines
        self.dependencies = dependencies
        self.migration_dependencies = migration_dependencies
        self.node = self.get_or_prepare_node(app_name, name)
        self.next: Optional[GraphNode] = None
        self.previous: Optional[GraphNode] = None

    @staticmethod
    def get_or_prepare_node(app_name, name) -> Node:
        """
        Return existing Node or object which is not created yet.

        :param app_name: target app
        :param name: name of migration
        :return: node with app_name=app_name, name=name
        """
        node_obj = Node(app_name=app_name, name=name)
        if node_obj.exists():
            node = node_obj.qs.filter(
                app_name=app_name, name=name).first()
            node_obj.pk = node.pk
            node_obj.created_at = node.created_at
        return node_obj

    def set_applied(self):
        """Queries django's migration table to determine whether node was applied before or not"""

        if self.node.is_applied:
            return

        recorder = MigrationRecorder(connections['default'])
        dependency_names = [d.split('.')[-1] for d in self.migration_dependencies]
        django_migrations = recorder.migration_qs.filter(
            app=self.node.app_name, name__in=dependency_names)
        if django_migrations.count() == len(dependency_names):
            self.node.apply()

    def apply(self) -> None:
        """
        Apply node.

        Calling this method will execute the routines within the node.
        If the migration is already applied do nothing.
        """
        if self.node.is_applied:
            return

        if self.prepare_migration_state():
            connection = connections['default']

            # Work out which apps have migrations and which do not
            executor = MigrationExecutor(connection)
            pre_migrate_state = executor._create_project_state(
                with_applied_migrations=True)
            current_state_apps = pre_migrate_state.apps
            with connection.schema_editor(atomic=True) as schema_editor:
                for routine in self.routines:
                    routine(
                        apps=current_state_apps,
                        schema_editor=schema_editor
                    )
        self.node.apply()

    def revert(self) -> None:
        """Reverts an applied migration node."""
        if not self.node.is_applied:
            return

        backup_node = Node(
            name=self.node.name,
            app_name=self.node.app_name
        )
        self.node.qs.get(pk=self.node.pk).delete()
        self.node = backup_node

    def prepare_migration_state(self):
        """Migrates the django migration graph until dependencies are applied"""
        # Get the database we're operating from
        connection = connections['default']

        # Work out which apps have migrations and which do not
        executor = MigrationExecutor(connection)
        pre_migrate_state = executor._create_project_state(
            with_applied_migrations=True)
        migration_graph_in_place = True

        if self.migration_dependencies:
            recorder = MigrationRecorder(connections['default'])
            applied_migrations = recorder.applied_migrations()
            unapplied_dependencies = list()
            for dependency in self.migration_dependencies:
                plan = dependency.split('.')
                plan = tuple([plan[-2], plan[-1]])
                if plan not in applied_migrations:
                    unapplied_dependencies.append(plan)

            if unapplied_dependencies:
                try:
                    plan = executor.migration_plan(unapplied_dependencies)
                    executor.migrate(
                        unapplied_dependencies,
                        plan=plan,
                        state=pre_migrate_state.clone()
                    )
                    migration_graph_in_place = True
                except NodeNotFoundError as ex:
                    for dep in unapplied_dependencies:
                        if not recorder.migration_qs.filter(
                                app=dep[0],
                                name__icontains=dep[1]
                        ).exists():
                            raise ex

            if not migration_graph_in_place:
                migration_graph_in_place = (
                    recorder
                    .migration_qs
                    .filter(app=self.node.app_name)
                    .order_by('-pk').first() in self.migration_dependencies
                )

        return migration_graph_in_place

    def append(self, node) -> None:
        """
        Append a node to the graph by attaching it to the last element.

        :param node: child to append
        """
        if not self.next:
            self.next = node
            node.previous = self
        else:
            self.next.append(node)

    @classmethod
    def from_struct(cls, app_name, obj) -> 'GraphNode':
        """
        Construct node from structured object.

        :param app_name: target application name
        :param obj: structured object to create from
        :return: representation as graph node
        """
        return cls(
            app_name,
            obj.name,
            obj.dependencies,
            obj.migration_dependencies,
            obj.routines
        )

    def __repr__(self) -> str:
        """Node representation."""
        date_str = self.node.created_at.isoformat() \
            if self.node.created_at else ""
        return f'{self.node.name}({date_str})'


class Graph:
    """Directed graph, imitates django's migration graph."""

    class MigrationNotFoundError(Exception):
        """Raised when non existing migration requested."""

        def __init__(self, name):
            super().__init__(f'Data migration "{name}" not found.')

    class EmptyGraphError(Exception):
        """Raised when trying to apply empty graph."""

        def __init__(self):
            super().__init__('Empty graph can\'t be applied.')

    def __init__(self, app_name: str) -> None:
        self.app_name = app_name
        self.base_node: Optional[GraphNode] = None

    def __repr__(self) -> str:
        """Representation of graph."""
        node = self.base_node
        out = str(self.base_node)
        while node.next:
            out += f'->{node.next}'
            node = node.next
        return out

    def push_back(self, child: GraphNode) -> None:
        """
        Append the graph with a new node, or set the first node as base node.

        :param child: node to append
        """
        if not self.base_node:
            self.base_node = child
        else:
            self.base_node.append(child)

    def get_node(self, name: str) -> GraphNode:
        """
        Getter for a node in the graph based on it's name.

        :raises Graph.MigrationNotFoundError: when base not available
        :param name: name of node
        :return: node within the graph
        """
        if name == 'zero':
            return self.base_node

        # no name = latest applied
        if not name:
            try:
                latest_node = Node.Node.objects.filter(
                    app_name=self.app_name).latest()
                name = latest_node.name
            except Node.Node.DoesNotExist:
                return self.base_node

        # search for node in tree with matching name
        node = self.base_node
        if node:
            if node.node.name == name:
                return node
            while node.next:
                if node.node.name == name:
                    return node

                node = node.next

        raise Graph.MigrationNotFoundError(name)

    def apply(self, name: Optional[str] = None,
              fail_silently: bool = False) -> None:
        """
        Apply the migration graph until (and including) given name.

        :raises Graph.EmptyGraphError(): on attempt on empty graph
        :param name: target migration name
        :param fail_silently: raise exception when applying empty graph
        """
        node: GraphNode = self.get_node(name)
        if not node:
            if not fail_silently:
                raise Graph.EmptyGraphError()
            return

        # revert or apply migrations depending on current state
        if (node.next and node.next.node.is_applied) or name == 'zero':
            self.revert_graph(node, name == 'zero')
        else:
            self.forward_graph(node, not bool(name))

    def forward_graph(self, node, complete: bool = True) -> None:
        """
        Apply graph based on nodes recursive.

        :param node: current node to apply
        :param complete: indicator to apply until last node
        """
        latest_node = node
        while (latest_node.previous
               and not latest_node.previous.node.is_applied):
            latest_node = latest_node.previous

        while latest_node:
            latest_node.apply()
            if not complete and latest_node == node:
                break
            latest_node = latest_node.next

    def revert_graph(self, node, complete: bool = False) -> None:
        """
        Reverts graph based on nodes recursive.

        :param node: current node to revert
        :param complete: indicator to revert whole graph
        """
        # walk down the tree
        latest_node = node
        while latest_node.next and latest_node.next.node.is_applied:
            latest_node = latest_node.next

        # revert until given point
        while latest_node:
            latest_node.revert()
            if not complete and latest_node == node:
                break
            latest_node = latest_node.previous

    @staticmethod
    def from_dir(app_name: str) -> 'Graph':
        """
        Generate graph from given app directory.

        :param app_name: name of app to generate graph of
        :return: Fully generated graph for requested app
        """
        app_conf = apps.get_app_config(app_name)
        dir_path = app_conf.path
        dir_path = os.path.join(dir_path, 'data_migrations')
        obj = Graph(app_name)
        files = [f for f in os.listdir(dir_path)
                 if os.path.isfile(os.path.join(dir_path, f))]
        for file in files:
            if file == '__init__.py':
                continue
            file = file.split('.')[0]
            module_name = f'{app_conf.module.__name__}.data_migrations.{file}'
            try:
                node = import_module(module_name).Node
            except AttributeError:
                continue
            obj.push_back(GraphNode.from_struct(app_name, node))
        return obj



from django.core.management.commands.makemigrations import Command as Migrate

from data_migration.services.file_generator import DataMigrationGenerator


class Command(Migrate):
    """Extended makemigrations command."""

    def add_arguments(self, parser):  # noqa D102
        parser.add_argument(
            '--data-only', action='store_true', dest='data_migration',
            help='Creates a data migration file.',
        )
        super().add_arguments(parser)

    def handle(self, *app_labels, **options):  # noqa D102
        create_empty_data_migration = options.get('data_migration')
        if create_empty_data_migration:
            is_dry_run = options.get('dry_run')
            return 'Generated files: ' + ', '.join([
                DataMigrationGenerator(
                    app,
                    readable_name=options.get('name'),
                    set_header=options.get('include_header', True),
                    empty=options.get('empty', False),
                    dry_run=is_dry_run
                ).file_name for app in app_labels])

        else:
            return super().handle(*app_labels, **options)



import logging
import os
from dataclasses import dataclass
from importlib import import_module
from typing import Optional, List

from django.apps import apps
from django.db import connections
from django.db.migrations.recorder import MigrationRecorder
from django.template import engines
from django.utils import timezone

from data_migration.helper import get_package_version_string
from data_migration.services.graph import GraphNode

current_dir = os.path.dirname(__file__)

log = logging.Logger(__file__)


@dataclass
class Routine:
    method: str
    module: str
    module_name: str
    file_path: str


class DataMigrationGenerator:
    migration_template = os.path.join(
        current_dir,
        'templates/migration.py.txt'
    )

    def __init__(self, app_name, readable_name: Optional[str] = None,
                 set_header: bool = True, empty: bool = False,
                 dry_run: bool = False, routines: List[Routine] = None,
                 migration_dependencies: List[str] = None) -> None:
        self.app_name = app_name
        if isinstance(self.app_name, list):
            self.app_name = self.app_name[0]
            log.warning(
                f'Initialize "{self.__class__.__name__}" using list,'
                f' will generate file for {self.app_name}.'
            )

        self.clean_app_name()
        root_dir = apps.get_app_config(self.app_name).path
        self.file_dir = os.path.join(root_dir, 'data_migrations')
        self.set_header = set_header
        self.empty = empty
        self.dry_run = dry_run
        self.routines: List[Routine] = routines
        if migration_dependencies is None:
            migration_dependencies = []
        self.migration_dependencies: List[str] = migration_dependencies
        self._gen_file(readable_name)

    def clean_app_name(self):
        self.app_name = self.app_name.replace('[\'', '').replace('\']', '')

    def _gen_file(self, readable_name: Optional[str] = None):
        empty_dir = True
        files = []
        if os.path.isdir(self.file_dir):
            files = [
                f for f in os.listdir(self.file_dir)
                if (f != '__init__.py'
                    and os.path.isfile(os.path.join(self.file_dir, f)))
            ]
            empty_dir = len(files) == 0
        elif not self.dry_run:
            # create directory
            os.mkdir(self.file_dir)

            file = open(os.path.join(self.file_dir, '__init__.py'), 'x')
            file.close()

        latest_filename = ''
        if empty_dir:
            # create first migration file
            self.file_name = os.path.join(
                self.file_dir,
                f'0001_{readable_name or "first"}.py'
            )
        else:
            latest_filename = sorted(files)[-1]
            latest_id = int(latest_filename[:4])

            # 0001_first.py or xxxx_auto.py
            name = readable_name or ("auto" if latest_id > 0 else "first")
            file_name = f'{self._get_id(latest_id+1)}_{name}.py'
            self.file_name = os.path.join(self.file_dir, file_name)

        if self.dry_run:
            return

        file = open(self.file_name, 'x')
        file.close()

        if self.empty:
            return

        template_kwargs = {
            'file_name': self.file_name.split('/')[-1].replace('.py', ''),
            'set_header': self.set_header,
            'date': timezone.now().strftime('%Y-%m-%d %H:%M'),
            'package': get_package_version_string(),
            'routines': self.routines,
            'migration_dependencies': self.migration_dependencies
        }

        if not empty_dir and latest_filename:
            template_kwargs.update({
                'dependencies': [latest_filename.replace('.py', '')]
            })

        recorder = MigrationRecorder(connections['default'])
        latest_migration = recorder.migration_qs.filter(
            app=self.app_name
        ).order_by('-applied').first()
        if latest_migration:
            template_kwargs['migration_dependencies'].append(
                f'{self.app_name}.{latest_migration.name}'
            )

        file = open(self.file_name, 'w')
        with open(self.migration_template, 'r') as input_file:
            file.write(self.render_template(
                input_file.read(), **template_kwargs)
            )
        file.close()

    @staticmethod
    def _get_id(index: int) -> str:
        out = str(index)
        while len(out) < 4:
            out = f'0{out}'
        return out

    @staticmethod
    def render_template(content, **context):
        django_engine = engines['django']
        template = django_engine.from_string(content)

        return template.render(context=context)

    def set_applied(self):
        file = self.file_name.split('.')[0].split('/')[-1]
        app_conf = apps.get_app_config(self.app_name)
        module_name = f'{app_conf.module.__name__}.data_migrations.{file}'

        try:
            node = import_module(module_name).Node
        except ModuleNotFoundError:
            return
        node = GraphNode.from_struct(self.app_name, node)
        node.set_applied()



from django.core.management.commands.squashmigrations import Command as Migrate


class Command(Migrate):
    """Alias for default `squashmigrations` command provided by Django."""

    def handle(self, **options):  # noqa D102
        return super().handle(**options)



from django.conf import settings
from django.apps import AppConfig

from data_migration.settings import internal_settings


class DataMigrationsConfig(AppConfig):
    name = 'data_migration'
    verbose_name = 'Django data migrations'

    def ready(self):
        internal_settings.update(settings)



from django.apps import apps
from django.core.management import CommandError
from django.core.management.commands.migrate import Command as Migrate

from data_migration.services.graph import Graph


class Command(Migrate):
    """
    Extended migrate command.

    Allows forward and backward migration of data/regular migrations
    """

    def add_arguments(self, parser):  # noqa D102
        parser.add_argument(
            '--data-only', action='store_true', dest='data_migration',
            help='Applies data migrations',
        )
        super().add_arguments(parser)

    def handle(self, *args, **options):  # noqa D102
        # extract parameters

        if options['app_label']:
            # Validate app_label.
            app_label = options['app_label']
            try:
                apps.get_app_config(app_label)
            except LookupError as err:
                raise CommandError(str(err))

            data_migrations = Graph.from_dir(app_label)
            data_migrations.apply(options.get('migration_name'))

        if not options['data_migration']:
            return super().handle(*args, **options)



import os

from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='django-data-migrations',
    version=os.getenv('VERSION'),
    packages=['data_migration'],
    author='Philipp Zettl',
    author_email='philipp.zett@godesteem.de',
    include_package_data=True,
    description='Extraction tool for data only django migrations',
    keywords='django,database migrations',
    url='https://github.com/philsupertramp/django-data-migration/',
    license='MIT',
    install_requires=[
        'django >= 2.2'
    ],
    long_description=read('README.rst'),
    long_description_content_type='text/x-rst',
    classifiers=[
        # 3 - Alpha
        # 4 - Beta
        # 5 - Production/Stable
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 2.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content"
    ]
)



class classproperty:
    """
    Decorator that converts a method with a single cls argument into a property
    that can be accessed directly from the class.
    """
    def __init__(self, method=None):
        self.fget = method

    def __get__(self, instance, cls=None):
        return self.fget(cls)


class AlreadyAppliedError(Exception):
    def __init__(self, node: 'Node.Node'):
        super().__init__(f'Node {node} already applied. Do not reapply them!')


class DatabaseError(Exception):
    def __init__(self, message):
        super().__init__(message)


class Node:
    """
    This is literally the same as django's MigrationRecorder
    """
    _node_model = None

    @classproperty
    def Node(cls):
        """
        bypass missing appconfig
        """
        if cls._node_model is None:
            from django.apps.registry import Apps
            from django.db import models

            class NodeClass(models.Model):
                app_name = models.CharField(max_length=255)
                name = models.CharField(max_length=255)
                created_at = models.DateTimeField()

                class Meta:
                    apps = Apps()
                    app_label = 'data_migration'
                    db_table = 'data_migrations'
                    get_latest_by = 'created_at'
                    constraints = [
                        models.UniqueConstraint(
                            fields=['app_name', 'name'],
                            name='unique_name_for_app'
                        )
                    ]

            cls._node_model = NodeClass
        return cls._node_model

    def __init__(self, name, app_name, *args, **kwargs):
        self.pk = kwargs.get('pk')
        self.name = name
        self.app_name = app_name
        self.created_at = kwargs.get('created_at')

    @property
    def is_applied(self):
        return bool(self.created_at and self.pk)

    def apply(self) -> None:
        if self.pk:
            raise AlreadyAppliedError(self)

        self.ensure_table()
        from django.utils import timezone
        self.created_at = timezone.now()

        obj = self.Node(
            name=self.name,
            app_name=self.app_name,
            created_at=self.created_at
        )
        obj.save()
        self.pk = obj.pk
        self.created_at = obj.created_at

    @property
    def qs(self):
        return self.Node.objects

    def exists(self):
        self.ensure_table()
        return self.qs.filter(
            app_name=self.app_name,
            name=self.name
        ).exists()

    @classmethod
    def flush(cls):
        return cls.get_qs().delete()

    @classmethod
    def get_qs(cls):
        node = cls('', '')
        node.ensure_table()
        return node.qs.all()

    def has_table(self):
        from django.db import connections
        with connections['default'].cursor() as cursor:
            tables = connections['default'].introspection.table_names(cursor)
        return self.Node._meta.db_table in tables

    def ensure_table(self):
        if self.has_table():
            return
        from django.db import connections, DatabaseError as DjDatabaseError
        # Make the table
        try:
            with connections['default'].schema_editor() as editor:
                editor.create_model(self.Node)
        except DjDatabaseError as ex:
            raise DatabaseError(
                f'Table "data_migrations" not creatable ({str(ex)}'
            )



import os
import sys
from typing import List, Dict

from data_migration.services.file_generator import DataMigrationGenerator,\
    Routine
from django.apps import apps
from django.core.management import call_command
from django.db.migrations import Migration
from django.db.migrations.executor import MigrationExecutor
from django.db import connection


MigrationPlan = List[Migration]
MigrationGraph = Dict[str, List[Migration]]


class Log:
    def __init__(self, out=sys.stdout):
        self.stdout = out
        self.log = ''

    def write(self, message):
        self.log += f'{message}\n'

    def flush(self):
        self.stdout.write(self.log)
        self.log = ''


log = Log()
sys.stdout = log


class MigrationFile:
    def __init__(self, fn: str, replacement_string: str, pos: int):
        self.file_name: str = fn
        self.replacement_string: str = replacement_string
        self.position: int = pos


class MigrationManager:
    def __init__(self, app_name: str, starting_point: str,
                 end_point: str, dry_run: bool = False):
        self.app_name: str = app_name
        self.start: str = starting_point
        self.end: str = end_point
        self.requires_action: bool = False
        self.migration_files_to_touch: List[MigrationFile] = []
        self.file_name = None
        self.new_file_name = None
        self.temp_file_name = None
        self.temp_directory = None
        self.dry_run = dry_run

    def load(self):
        self.file_name = self.gen_file_name()
        self.new_file_name = self.gen_new_file_name()
        self.temp_file_name = self.gen_temp_file_name()
        self.temp_directory = '/'.join(self.file_name.split('/')[:-1] + ['temp'])

    def __repr__(self):
        return self.file_name

    @property
    def app_path(self) -> str:
        return apps.get_app_config(self.app_name).path

    def gen_file_name(self) -> str:
        return f'{self.app_path}/migrations/{self.start}' \
               f'_squashed_{self.end}.py'

    def gen_temp_file_name(self) -> str:
        return f'{self.app_path}/migrations/temp/{self.start}' \
               f'_squashed_{self.end}.py'

    def gen_new_file_name(self) -> str:
        return f'{self.app_path}/migrations/' \
               f'{self.start.split("_")[0]}' \
               f'_squashed_{self.end.split("_")[0]}.py'

    def _parse_generated_file(self):
        parse_start = False
        with open(self.temp_file_name, 'r') as file:
            char_count = 0
            line = "–"
            replacement_string = '# RunPython operations to refer ' \
                                 'to the local versions:'
            while line != "":
                line = file.readline()
                if replacement_string in line:
                    parse_start = True
                    self.requires_action = True
                elif parse_start and '#' in line:
                    partial_path = line.replace('# ', '').replace('\n', '')
                    migration_path = f'{partial_path.replace(".", "/")}.py'
                    self.migration_files_to_touch.append(
                        MigrationFile(
                            migration_path,
                            partial_path,
                            char_count
                        )
                    )

                char_count += len(line)

    def run(self):
        self.load()

        if self.dry_run:
            log.write(f'Squashing {self.app_name}: '
                      f'"{self.start}"–"{self.end}"')
            return

        call_command(
            'django_squashmigrations',
            self.app_name,
            self.start,
            self.end,
            '--no-input',
            stdout=log
        )

    def post_processing(self):
        """
        post processing of squashed migration files,
        Drops part of django's squashmigration auto generated
        comments/commands that makes migration files unusable.
        """
        if self.dry_run:
            return

        self.load()
        try:
            os.mkdir(self.temp_directory)
        except FileExistsError:
            pass
        os.rename(self.file_name, self.temp_file_name)
        self._parse_generated_file()
        if not self.requires_action:
            return

        with open(self.temp_file_name, 'r') as read_file,\
             open(self.new_file_name, 'w') as write_file:
            stack_open = False
            line = "–"
            while line != "":
                line = read_file.readline()
                if 'RunPython(' in line:
                    stack_open = True
                    continue

                if stack_open and ')' in line:
                    stack_open = False
                    continue

                found = False
                for elem in self.migration_files_to_touch:
                    if elem.replacement_string in line:
                        found = True
                        index = line.find(elem.replacement_string)
                        next_space = line.find(
                            ' ',
                            index + len(elem.replacement_string)
                        )
                        if next_space == -1:
                            next_space = line.find(
                                ',',
                                index + len(elem.replacement_string)
                            )
                        if next_space == -1:
                            next_space = line.find(
                                '\n',
                                index + len(elem.replacement_string)
                            )
                        elem.replacement_string = line[index:next_space]
                        break

                if stack_open or 'code=' in line or '#' in line or found:
                    continue

                write_file.write(line)

        os.remove(self.temp_file_name)

    def process_data_migrations(self):
        """
        Generated data_migrations based on processed files.
        """
        for elem in self.migration_files_to_touch:
            module_name = elem.replacement_string.split(".")[-2]
            module = '.'.join(elem.replacement_string.split('.')[:-1])
            generator = DataMigrationGenerator(
                self.app_name, module_name,
                routines=[
                    Routine(
                        method=elem.replacement_string.split('.')[-1],
                        module=module,
                        module_name=f'mig_{module_name}',
                        file_path=elem.file_name,
                    )
                ],
                migration_dependencies=[module.replace('.migrations', '')],
                dry_run=self.dry_run,
            )
            generator.set_applied()


class MigrationSquash:
    def __init__(self, _loa: List[str], dry_run: bool = False):
        self.executor: MigrationExecutor = MigrationExecutor(connection)
        self.plan: MigrationPlan = []
        self.graph: MigrationGraph = {}
        self.get_migration_graph()
        self.parse_plan()
        self.list_of_apps = _loa
        self.dry_run = dry_run

    @property
    def log(self):
        return log

    def get_migration_graph(self):
        django_plan = self.executor.migration_plan(
            self.executor.loader.graph.leaf_nodes(),
            clean_start=True
        )
        self.plan = [i[0] for i in django_plan]

    def parse_plan(self):
        indices: Dict[str, int] = dict()
        last_key: str = ''
        for migration in self.plan:
            app = migration.app_label
            if app not in indices:
                indices[app] = 0
                self.graph[f'{app}_0'] = [migration]
            elif last_key == app:
                self.graph[f'{app}_{indices[app]}'].append(migration)

            elif len(list(self.graph.keys())) > 0 \
                    and self.graph[list(self.graph.keys())[-1]] != app:
                indices[app] += 1
                self.graph[f'{app}_{indices[app]}'] = [migration]
            else:
                self.graph[f'{app}_{indices[app]}'].append(migration)
            last_key = app

    def squash(self):
        for index, street in enumerate(self.graph.keys()):
            app_name: str = self.graph[street][0].app_label
            if len(self.graph[street]) > 1 and app_name in self.list_of_apps:
                from_id = self.graph[street][0].name
                to_id = self.graph[street][-1].name

                mig = MigrationManager(app_name, from_id, to_id, self.dry_run)
                mig.run()
                mig.post_processing()
                mig.process_data_migrations()
